from .dottedDict import ensureDottedAccess, DottedDict
from .path import (MULTI_LEVEL_WILDCARD, SINGLE_LEVEL_WILDCARD, SPLITCHAR,
                   containsWildcards, patternIsValid)


def generateResult(res=DottedDict({})):
    """ internal helper to generate the defined result.

    Args:
        res (dotted_dict, optional): The Result. Defaults to dotted_dict({}).

    Returns:
        dotted_dict: The Result
    """
    defaultResult = DottedDict({
        'affected': False,
        'affectedByChild': False,
        'affectedByParent': False,
        'affectedOnSameLevel': False,
        'containsWildcards': False,
        'patternToExtractData': False,
        'patternLengthComparedToPathLength': '=',
        'pathToExtractData': False
    })

    defaultResult.update(res)
    defaultResult.affected = defaultResult.affectedByChild or defaultResult.affectedByParent or defaultResult.affectedOnSameLevel

    return defaultResult


def comparePatternAndPath(pathPattern: str, contentPath: str, _options=None):
    """ Matches the given path, with the pattern and determines, if the path might affect the given pattern.

        example: path = "a/b/c"; pattern = "a/#"; => totalPath = "a/b/c"; diffPath = "b/c"

    Args:
        pathPattern (str): The pattern to test
        contentPath (str): The path to use as basis
        options (dotted_dict, optional): _description_. Defaults to dotted_dict({'matchTopicsWithoutWildcards': False}).

    Returns:
        dotted_dict: The Result
    """
    options = ensureDottedAccess({'matchTopicsWithoutWildcards': False})
    options.update(ensureDottedAccess(_options))

    if containsWildcards(contentPath):
        raise Exception(
            "The Path is invalid. The path should not contain pattern-related chars '#' or '+'.")
    if not patternIsValid(pathPattern):
        raise Exception('The Pattern is invalid.')
    if not patternIsValid(contentPath):
        raise Exception('The Path is invalid.')

    _containsWildcards = containsWildcards(pathPattern)
    _patternSegments = pathPattern.split(SPLITCHAR)
    _contentPathSegments = contentPath.split(SPLITCHAR)
    _patternLength = len(_patternSegments)
    _contentPathLength = len(_contentPathSegments)

    # Define the Char for the comparer
    patternLengthComparedToPathLength = '='
    if _patternLength > _contentPathLength:
        patternLengthComparedToPathLength = '>'
    elif _patternLength < _contentPathLength:
        patternLengthComparedToPathLength = '<'

    # If both, the pattern and the path are equal => return the result.
    if pathPattern == contentPath:
        return generateResult(
            DottedDict(
                {
                    'affectedOnSameLevel': True,
                    'pathToExtractData': contentPath,
                    'patternLengthComparedToPathLength': patternLengthComparedToPathLength
                }
            )
        )

    # If the Path is not realy defined.
    if contentPath == '':
        return generateResult(
            DottedDict({
                'affectedByParent': True,
                'patternToExtractData': pathPattern if _containsWildcards else False,
                'pathToExtractData': False if _containsWildcards else pathPattern,
                'patternLengthComparedToPathLength': '>',
                'containsWildcards': _containsWildcards
            })
        )
    if pathPattern == '':
        return generateResult(
            DottedDict({
                'affectedByChild': True,
                'pathToExtractData': '',
                'patternLengthComparedToPathLength': '<'
            })
        )
    if options.matchTopicsWithoutWildcards:
        if contentPath.startswith(pathPattern):
            # Path is longer then the Pattern;
            # => A Change is performed by "Child",
            if _containsWildcards:
                return generateResult(
                    DottedDict({
                        'affectedByChild': True,
                        'pathToExtractData': contentPath,
                        'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
                        'containsWildcards': _containsWildcards
                    })
                )
            else:
                return generateResult(
                    DottedDict({
                        'affectedByChild': True,
                        'pathToExtractData': pathPattern,
                        'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
                        'containsWildcards': _containsWildcards
                    })
                )
        elif pathPattern.startswith(contentPath):
            # Pattern is longer then the path;
            # => A Change might be initated by
            # the super element

            # The PathToExtractData is allways false, if the path is smaller then the
            # pattern

            if _containsWildcards:
                return generateResult(
                    DottedDict({
                        'affectedByParent': True,
                        'patternToExtractData': pathPattern,
                        'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
                        'containsWildcards': _containsWildcards
                    })
                )
            else:
                return generateResult(
                    DottedDict({
                        'affectedByParent': True,
                        'pathToExtractData': pathPattern,
                        'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
                        'containsWildcards': _containsWildcards
                    })
                )

    partialPath = ''
    i = 0

    # Iterate over the Segments.
    while i < _patternLength:
        # Store the current Pattern Segment
        currentPattern = _patternSegments[i]

        # We need to know, if there is SINGLE_LEVEL_WILDCARD or MULTI_LEVEL_WILDCARD
        # there fore we will extract the Wildlevels.
        patternChar = currentPattern[0]
        currentPath = _contentPathSegments[i] if i < _contentPathLength else None
        i = i + 1
        if currentPath is None:
            # Our Pattern is larger then our contentPath.
            # So we dont know, whether we will get some
            # data. Therefore we have to perform a query
            # later ==> Set The Path / Pattern.

            if _containsWildcards:
                # But we contain Patterns.
                # So we are not allowed to build a
                # diff object.
                return generateResult(
                    DottedDict({
                        'affectedByParent': True,
                        'patternToExtractData': pathPattern,
                        'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
                        'containsWildcards': _containsWildcards
                    })
                )
            elif patternLengthComparedToPathLength == ">":
                return generateResult(
                    DottedDict({
                        'affectedByParent': True,
                        'patternToExtractData': pathPattern,
                        'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
                        'containsWildcards': _containsWildcards
                    })
                )
            else:
                raise Exception('Implementation Error! This should not happen')
        elif currentPath == currentPattern:
            # The Patterns Match
            # We now store the correct path of our segment.
            partialPath = f'{partialPath}{SPLITCHAR}{currentPath}' if len(
                partialPath) > 0 else currentPath
        elif patternChar == MULTI_LEVEL_WILDCARD:

            # We know, that MULTI_LEVEL_WILDCARDs are only at the end of the
            # pattern. So it might happen, that:
            # a)   our length of the pattern is the same length as the content path
            # b)   our length of the pattern is smaller then length as the content path
            #
            # Our statement before alread tested, that either case a) or b) fits. Otherwise
            # another ifstatement is valid and we wont enter this statement
            # here.

            # # We add the segment to testedCorrectPath
            # testedCorrectPath = testedCorrectPath.length > 0 ?
            # `{testedCorrectPath}{SPLITCHAR}{currentPath}` : currentPath;

            if patternLengthComparedToPathLength == '=':
                # Case a)
                return generateResult(
                    DottedDict(
                        {
                            'affectedOnSameLevel': True,
                            'pathToExtractData': contentPath,
                            'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
                            'containsWildcards': _containsWildcards
                        }
                    )
                )
            elif patternLengthComparedToPathLength == '<':
                # Case b)
                return generateResult(DottedDict(
                    {
                        'affectedByChild': True,
                        'pathToExtractData': contentPath,
                        'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
                        'containsWildcards': _containsWildcards
                    }
                ))
            else:
                raise Exception('Implementation Error!')
        elif patternChar == SINGLE_LEVEL_WILDCARD:
            # Store the correct path.
            partialPath = f'{partialPath}{SPLITCHAR}{currentPath}' if len(
                partialPath) > 0 else currentPath
        elif patternChar != SINGLE_LEVEL_WILDCARD and currentPattern != currentPath:
            return generateResult(DottedDict({
                'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
                'containsWildcards': _containsWildcards
            }))
    diff = contentPath[len(partialPath):]
    return generateResult(DottedDict({
        'affectedOnSameLevel': len(diff) == 0,
        'affectedByChild': len(diff) >= 1,
        'pathToExtractData': partialPath,
        'patternLengthComparedToPathLength': patternLengthComparedToPathLength,
        'containsWildcards': _containsWildcards
    }))



import asyncio

from nope.helpers import Emitter, generateId, formatException, ensureDottedAccess, \
    EXECUTOR
from nope.logger import defineNopeLogger
from nope.observable import NopeObservable


class Bridge:

    def __init__(self, _id=generateId(), logger=False):

        def getter(storedValue):
            for data in self._layers.values():
                if (data.considerConnection and not data.layer.connected.
                        getContent()):
                    return False
            return True

        self.considerConnection = True
        self.allowServiceRedundancy = False
        self._internalEmitter = Emitter()
        self._callbacks = dict()
        self._layers = dict()
        self._id = _id
        self._logger = defineNopeLogger(logger, 'nope.bridge')
        self._useInternalEmitter = True

        self.connected = NopeObservable()
        self.connected.setContent(False)
        self.connected.getter = getter

        self._subscribedEvents = dict()

    @property
    def receivesOwnMessages(self):
        for layer in self._layers.values():
            if not layer.layer.receivesOwnMessages:
                return False
        return True

    @property
    def id(self):
        return self._id

    async def on(self, eventName: str, cb):
        return await self._on(eventName, lambda data: cb(ensureDottedAccess(data)))

    async def emit(self, eventName: str, data, **kwargs):
        return await self._emit(eventName, None, ensureDottedAccess(data))

    def detailListeners(self, t, listeners):
        raise Exception('Method not implemented.')

    async def dispose(self):
        for item in self._layers.values():
            await item.layer.dispose()

    def _checkInternalEmitter(self):
        self._useInternalEmitter = True
        for layer in self._layers.values():
            if layer.layer.receivesOwnMessages:
                self._useInternalEmitter = False
                break

    async def _subscribeToCallback(self, layer, event, forwardData):
        if forwardData:
            await layer.on(event, lambda data: EXECUTOR.callParallel(self._emit, event, layer, data))
        else:
            await layer.on(event, lambda data: self._internalEmitter.emit(event, data))

    async def _on(self, event, cb):

        # if required we will add a logger for the events.
        if self._logger and event != 'StatusChanged':
            def debug_callback(data):
                if self._logger:
                    self._logger.debug(f'received "{event}", data={data}')
                    self._logger.debug(f'subscribe to "{event}"')

            self._internalEmitter.on(event, debug_callback)

        self._internalEmitter.on(event, cb)

        # We now store the callback.
        if event not in self._callbacks:
            self._callbacks[event] = [cb]
        else:
            self._callbacks[event].append(cb)

        promises = []

        # Now we try to add the callback to the connected layers.
        for data in self._layers.values():
            if data.layer.connected.getContent():
                promises.append(
                    self._subscribeToCallback(
                        data.layer, event, data.forwardData))

        # Now wait for every Layer.
        if promises:
            await asyncio.gather(*promises)

    async def _emit(self, event, toExclude, dataToSend=None, force=False):
        if self._logger and event != 'StatusChanged':
            self._logger.debug(f'emitting {str(event)} {str(dataToSend)}')
        if self._useInternalEmitter or force:
            self._internalEmitter.emit(event, dataToSend)

        # Collect all events
        promises = []

        for data in self._layers.values():
            if data.layer != toExclude and data.layer.connected.getContent():

                async def emitOnLayer():
                    try:
                        await data.layer.emit(event, dataToSend)
                    except Exception as error:
                        if self._logger:
                            self._logger.error(
                                'failed to emit the event "{event}"')
                            self._logger.error(error)
                        else:
                            print(formatException(error))

                promises.append(emitOnLayer())

        # Now wait for all Layers to emit
        if promises:
            await asyncio.gather(*promises)

    async def addCommunicationLayer(self, layer, forwardData=False, considerConnection=False):
        if layer.id not in self._layers:

            def onConnected(*args):
                self.connected.forcePublish()

            self._layers[layer.id] = ensureDottedAccess({
                'layer': layer,
                'considerConnection': considerConnection,
                'forwardData': forwardData
            })
            layer.connected.subscribe(onConnected)

            await layer.connected.waitFor()

            for event, cbs in self._callbacks.items():
                for callback in cbs:
                    layer.on(event, callback)
                    self._checkInternalEmitter()

    async def removeCommunicationLayer(self, layer):
        if layer.id in self._layers:
            self._layers.pop(layer.id)
            self._checkInternalEmitter()



from copy import deepcopy

from .hashable import hdict
from dataclasses import is_dataclass, asdict

DEFAULT_METHODS = dir(hdict)


class DottedDict(hdict):
    """dot.notation access to dictionary attributes. it is although hashable. but than it can not be edited."""

    def __getattr__(self, key):
        if key in DEFAULT_METHODS:
            return getattr(self, key)
        elif key == "__hash":
            return hdict.__getattr__(self, key, value)
        else:
            return self.get(key)

    def __setattr__(self, key, value):
        if key in DEFAULT_METHODS:
            raise Exception("This would overwrite the default behavior")
        elif key == "__hash":
            return hdict.__setattr__(self, key, value)
        hdict.__setitem__(self, key, value)

    def __delattr__(self, key):
        if key in DEFAULT_METHODS:
            raise Exception("This would overwrite the default behavior")
        elif key == "__hash":
            return hdict.__delattr__(self, key, value)
        hdict.__delitem__(self, key)

    def copy(self):
        # Manually copy the contained Dict.
        cp = {}
        for k, v in self.items():
            if hasattr(v, "copy") and callable(v.copy):
                cp[deepcopy(k)] = v.copy()
            else:
                cp[deepcopy(k)] = deepcopy(v)
        # use that to create a copy.
        return DottedDict(cp)

    def __repr__(self):
        return dict.__repr__(self)


# DottedDict = hashable(DottedDict)


class NoneDottedDict(DottedDict):
    """dot.notation access to dictionary attributes"""

    def __getattr__(self, key):
        if key in DEFAULT_METHODS:
            return getattr(self, key)
        else:
            try:
                return self.get(key)
            except KeyError:
                return None

    def __getitem__(self, item):
        try:
            return hdict.__getitem__(self, item)
        except KeyError:
            return None

    def copy(self):
        # Manually copy the contained Dict.
        cp = {}
        for k, v in self.items():
            if hasattr(v, "copy") and callable(v.copy):
                cp[deepcopy(k)] = v.copy()
            else:
                cp[deepcopy(k)] = deepcopy(v)
        # use that to create a copy.
        return NoneDottedDict(cp)


# NoneDottedDict = hashable(NoneDottedDict)

def _convertItem(item, useNoneAsDefaultValue: bool):
    """ Helper that converts the items to the corresponding type

    Args:
        item: The item to return

    Returns:
        any: the adapted item
    """
    if type(item) in (dict, DottedDict, NoneDottedDict):
        return convertToDottedDict(item, useNoneAsDefaultValue)
    elif is_dataclass(item):
        return convertToDottedDict(asdict(item))
    elif type(item) in (list, set):
        new_list = list()
        for i in item:
            new_list.append(_convertItem(i, useNoneAsDefaultValue))
        return new_list
    else:
        return item


def convertToDottedDict(d: dict | DottedDict |
                        NoneDottedDict, useNoneAsDefaultValue=True):
    """ Converts a dict to a dotted dict. Although, ensures,

    Args:
        d (dict): The dictionary.
        useNoneAsDefaultValue (bool): Flag to enable the default value 'none' instead of an KeyError
    """
    ret = NoneDottedDict() if useNoneAsDefaultValue else DottedDict()

    if is_dataclass(d):
        d = asdict(d)

    if isinstance(d, (dict, DottedDict, NoneDottedDict)):
        for k, v in d.items():
            ret[k] = _convertItem(v, useNoneAsDefaultValue)

    return ret


def ensureDottedAccess(d, useNoneAsDefaultValue=True):
    """ Ensure the given item is a dotted dict.

    Args:
        d (any): Item to convert
        useNoneAsDefaultValue (bool): A flag to enable default-nones.
    Returns:
        DottedDict: _description_
    """
    if d is None:
        return DottedDict()

    elif not isinstance(d, DottedDict):
        return convertToDottedDict(d, useNoneAsDefaultValue)

    return d



# @author Martin Karkowski
# @email m.karkowski@zema.de
# @desc [description]

import asyncio
import threading
import time
from asyncio import iscoroutinefunction

from .asyncHelpers import getOrCreateEventloop


class setInterval:
    """ Starts an interval, on which the function will be called.
    """

    def __init__(self, func, interval, *args):
        self.interval = interval / 1000
        self._is_coroutine = iscoroutinefunction(func)
        self._loop = None if not self._is_coroutine else getOrCreateEventloop()
        self.func = func
        self.args = args
        self.stop_event = threading.Event()
        self._thread = threading.Thread(target=self.__run)
        self._thread.start()

    def __run(self):
        next_time = time.time() + self.interval
        if self._is_coroutine:
            loop = getOrCreateEventloop()
            asyncio.set_event_loop(loop)

            while not self.stop_event.wait(next_time - time.time()):
                next_time += self.interval
                asyncio.run_coroutine_threadsafe(self.func(*self.args))

        else:
            while not self.stop_event.wait(next_time - time.time()):
                next_time += self.interval
                self.func(*self.args)

    def cancel(self):
        self.stop_event.set()


def setTimeout(func, msec, *args):
    """ Calls the function delayed.
    """
    if iscoroutinefunction(func):
        def func_wrapper():
            loop = getOrCreateEventloop()
            asyncio.set_event_loop(loop)
            asyncio.run_coroutine_threadsafe(func(*args))

        t = threading.Timer(msec / 1000.0, func_wrapper)
        t.start()
        return t
    else:
        def func_wrapper():
            func(*args)

        t = threading.Timer(msec / 1000.0, func_wrapper)
        t.start()
        return t



#!/usr/bin/env python

import logging
import sys

DEBUG = logging.DEBUG
INFO = logging.INFO
ERROR = logging.ERROR
WARNING = logging.WARNING
CRITICAL = logging.CRITICAL

LEVELS = {'info': INFO, 'debug': DEBUG, 'warn': WARNING, 'error': ERROR}

_LOGGERS = {}


class ColorizedFormatter(logging.Formatter):
    """Logging colored formatter, adapted from https://stackoverflow.com/a/56944256/3638629"""

    grey = '\x1b[38;21m'
    blue = '\x1b[38;5;39m'
    yellow = '\x1b[38;5;226m'
    red = '\x1b[38;5;196m'
    bold_red = '\x1b[31;1m'
    reset = '\x1b[0m'

    def __init__(self, fmt):
        super().__init__()
        self.fmt = fmt
        self.FORMATS = {
            logging.DEBUG: self.grey + self.fmt + self.reset,
            logging.INFO: self.blue + self.fmt + self.reset,
            logging.WARNING: self.yellow + self.fmt + self.reset,
            logging.ERROR: self.red + self.fmt + self.reset,
            logging.CRITICAL: self.bold_red + self.fmt + self.reset
        }

    def format(self, record):
        log_fmt = self.FORMATS.get(record.levelno)
        formatter = logging.Formatter(log_fmt)
        return formatter.format(record)


def getNopeLogger(name: str, level=logging.DEBUG):
    """ Helper to return a specific logger.

    """
    global _LOGGERS

    if name not in _LOGGERS:

        nameToUse = name
        length = 30

        if len(nameToUse) > length:
            nameToUse = nameToUse[:length - 3] + "..."

        _logger = logging.getLogger(nameToUse)
        # Define  a Logging Format

        _format = ColorizedFormatter(
            '%(asctime)s - %(levelname)-8s - %(name)-' + str(length) + 's - %(message)s')

        # Create Console Output
        _handler = logging.StreamHandler(sys.stdout)
        # Add the Format to the Handler
        _handler.setFormatter(_format)
        # Set Loglevel to the Desired One.
        _handler.setLevel(level)

        # Finally add the Handler to the Logger:
        _logger.addHandler(_handler)

        # Set the Log Level of the Logger.
        _logger.setLevel(level)

        _LOGGERS[name] = _logger

    return _LOGGERS[name]


def defineNopeLogger(param, default_name: str):
    """ Helper to define a Logger. The parameter 'param' is either a logger, None, or a Logger-Level

    """
    if not param:
        return None
    if isinstance(param, str):
        return getNopeLogger(default_name, LEVELS[param])
    return param



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

import sys

from .run import run_cli
from .scan import scan_cli
from .logo import renderLogo

help_message = '''
usage: nope-py [-h] mode

NoPE - Command Line Interface. Please select the option you want.

positional arguments:
    mode        mode of the cli-tool. Possible Options "run", "conf"

optional arguments:
    -h, --help  show this help message and exit
'''


def main_cli():
    if len(sys.argv) >= 2:
        if (sys.argv[1] == "run"):
            renderLogo()
            run_cli(True)
        elif (sys.argv[1] == "conf"):
            scan_cli(True)
        else:
            print(help_message)
    else:
        print(help_message)



from ..eventEmitter import NopeEventEmitter
from ..helpers import DottedDict, extractUniqueValues, transformDict, determineDifference
from ..observable import NopeObservable


class MergeData:

    def __init__(self, originalData, _extractData):
        self.originalData = originalData
        self._extractData = _extractData
        self.onChange = NopeEventEmitter()
        self.data = NopeObservable()
        self.data.setContent([])
        self.error = False

        self.update(force=True)

    def update(self, data=None, force=False):
        if data is not None:
            self.originalData = data
        after_adding = self._extractData(self.originalData)
        diff = determineDifference(self.data.getContent(), after_adding)
        if force or ((len(diff.removed) > 0) or len(diff.added) > 0):
            self.data.setContent(list(after_adding))
            self.onChange.emit(DottedDict(
                {'added': list(diff.added), 'removed': list(diff.removed)}))

    def dispose(self):
        self.data.dispose()
        self.onChange.dispose()


class DictBasedMergeData(MergeData):

    def __init__(self, originalData, _path='', _pathKey=None):
        def callback(m):
            return extractUniqueValues(m, _path, _pathKey)

        self._path = _path
        self._pathKey = _pathKey
        self.amountOf = dict()
        self.simplified = dict()
        self.keyMapping = dict()
        self.keyMappingreverse = dict()
        self.conflicts = dict()
        self.orgKeyToExtractedValue = dict()
        self.extracted_key = []
        self.extracted_value = []

        super().__init__(originalData, callback)

    def update(self, data=None, force=False):
        if data is not None:
            self.originalData = data
        result = transformDict(self.originalData, self._path, self._pathKey)

        self.simplified = result.extracted_map
        self.amountOf = result.amountOf
        self.keyMapping = result.keyMapping
        self.keyMappingreverse = result.keyMappingreverse
        self.conflicts = result.conflicts
        self.orgKeyToExtractedValue = result.orgKeyToExtractedValue
        self.extracted_key = [*self.simplified.keys()]
        self.extracted_value = [*self.simplified.values()]

        super().update(data, force)



# /**
#  * @author Martin Karkowski
#  * @email m.karkowski@zema.de
#  * @create date 2021-06-14 08:16:07
#  * @modify date 2021-06-14 08:35:51
#  * @desc [description]
#  */

import asyncio
import threading
from asyncio.coroutines import iscoroutinefunction


class FuncThread(threading.Thread):

    def __init__(self, target: callable, *args, **kwargs):
        threading.Thread.__init__(self)

        self._target = target
        self._args = args
        self._kwargs = kwargs

    def run(self):
        self._target(*self._args)


def offload_function_to_thread(func, loop):
    """ Helper Function to perform a Method in a separate Thread,
        Therefore you will receive an async function which will be
        called in the background.
    """

    loop = asyncio.get_event_loop()

    promise = loop.create_future()

    if iscoroutinefunction(func):
        # Not implemented Jet
        raise TypeError("You only are allowed to offload sync-functions")
    else:
        def _parallel(*args, **kwargs):
            try:
                result = func(*args, **kwargs)
                promise.set_result(result)
            except Exception as E:
                promise.set_exception(E)

        async def _func(*args, **kwargs):
            _main_thread = FuncThread(target=_parallel, *args, **kwargs)
            _main_thread.run()
            return await promise

        return _func



#!/usr/bin/env python
# @author Martin Karkowski
# @email m.karkowski@zema.de

import logging

from .nopePackkageLoader import NopePackageLoader
from ..dispatcher import getDispatcher


class NopePackageLoaderSingleton(object):
    _instance = None

    def __new__(cls, **settings):
        if cls._instance is None:
            cls._instance = NopePackageLoader(
                getDispatcher(settings, settings),
                **settings
            )
        # Put any initialization here.
        return cls._instance


def getPackageLoader(**settings):
    """ helper function, which will be used to create a package-loader
    """
    # Create the Dispatcher Instance.
    loader = NopePackageLoaderSingleton(**settings)
    return loader



#!/usr/bin/env python
# Author: Martin Karkowski


from setuptools import setup

try:
    long_description = open("./README.md").read()
except BaseException:
    long_description = """NOPE-Python Backend. A Generic Backend for Python"""

if __name__ == "__main__":
    setup(name="nope_py",
          version="1.7.8rc3",
          description="NOPE-Lib represents the No Programming Environment Library for python.",
          long_description=long_description,
          long_description_content_type="text/markdown",
          author="Martin Karkowski",
          author_email="m.karkowski@zema.de",
          maintainer="Martin Karkowski",
          maintainer_email="m.karkowski@zema.de",
          install_requires=['paho-mqtt',
                            'python-socketio[asyncio_client]',
                            'psutil'],
          url="https://github.com/ZeMA-gGmbH/NoPE-PY.git",
          packages=["nope",
                    "nope.cli",
                    "nope.communication",
                    "nope.communication.layers",
                    "nope.decorators",
                    "nope.demo",
                    "nope.demo.instances",
                    "nope.demo.plugins",
                    "nope.dispatcher",
                    "nope.dispatcher.baseServices",
                    "nope.dispatcher.connectivityManager",
                    "nope.dispatcher.core",
                    "nope.dispatcher.instanceManager",
                    "nope.dispatcher.rpcManager",
                    "nope.eventEmitter",
                    "nope.helpers",
                    "nope.observable",
                    "nope.logger",
                    "nope.loader",
                    "nope.merging",
                    "nope.modules",
                    "nope.plugins",
                    "nope.pubSub",
                    "nope.types"
                    ],
          package_data={'': ['VERSION']},
          include_package_data=True,
          classifiers=[
              #   "Development Status :: 4 - Production/Stable",
              "Operating System :: OS Independent",
              "Programming Language :: Python",
              "Topic :: Scientific/Engineering",
              "Topic :: Software Development :: Libraries :: Python Modules",
          ],
            entry_points={
                'console_scripts': [
                    'nope-py = nope.cli:main_cli',
                ],
            },
          )



import asyncio
import time
from asyncio import sleep

import pytest

from nope import getLayer, NopeRpcManager, formatException, getTimestamp, EXECUTOR

IN_PY_TEST = False


@pytest.fixture
def event_loop():
    global IN_PY_TEST
    IN_PY_TEST = True
    loop = asyncio.new_event_loop()
    EXECUTOR.assignLoop(loop)
    yield loop
    loop.close()


async def main():
    manager = NopeRpcManager({
        "communicator": await getLayer("event"),
        "logger": False,
    }, lambda *args: "test", "test")

    await manager.ready.waitFor()

    async def hello(name: str) -> str:
        return f"Hello {name}!"

    async def delayed(name: str) -> str:
        await sleep(1)
        return await hello(name)

    manager.registerService(hello, {
        "id": "hello"
    })

    manager.registerService(delayed, {
        "id": "delayed"
    })

    await sleep(1.5)

    assert "hello" in manager.services.extracted_key, "Failed to register the services"
    assert "delayed" in manager.services.extracted_key, "Failed to register the services"

    # Try calling the Service
    res = await manager.performCall("hello", ["Pytest"])
    assert res == "Hello Pytest!"

    try:
        res = await manager.performCall("delayed", ["Pytest"], {"timeout": 2000})
        assert res == "Hello Pytest!"
    except Exception as error:
        print("HERE", error)

    try:
        res = await manager.performCall("delayed", ["Pytest"], {"timeout": 200})
        assert res == "Hello Pytest!"
    except Exception as error:
        print(formatException(error))

    start = getTimestamp()
    res = await manager.performCall(["delayed"] * 5, ["Pytest"])
    end = getTimestamp()

    if not IN_PY_TEST:
        delta = []
        bench = 100000

        start = time.process_time()
        for i in range(bench):
            # s = time.process_time()
            await manager.performCall("hello", ["Benchmark"])
            # delta = (time.process_time() - s) * 1000
            # print(i, "mainloop:", delta, "[ms]")
        end = time.process_time()
        delta = end - start
        print(delta, bench)
        time_per_call = delta * 1000 / bench
        qps = 1000.0 / time_per_call

        print(
            "NOPE    ",
            round(
                time_per_call,
                5),
            "[ms] ->",
            round(
                qps,
                5),
            "[R/s]")

        start = getTimestamp()
        for i in range(bench):
            await hello("delayed")
        end = getTimestamp()
        delta = end - start
        time_per_call = delta / bench
        qps = 1000.0 / time_per_call

        print(
            "ORIGINAL",
            round(
                time_per_call,
                5),
            "[ms] ->",
            round(
                qps,
                5),
            "[R/s]")

    # await manager.dispose()


if __name__ == "__main__":
    EXECUTOR.loop.run_until_complete(main())
    # EXECUTOR.loop.run_forever()

# yappi.set_clock_type("CPU")
# with yappi.run():
#     EXECUTOR.loop.run_until_complete(main())
# yappi.get_func_stats().print_all()



from nope.communication.bridge import Bridge
from nope.communication.layers import MQTTLayer, IoSocketClientLayer
from nope.helpers import generateId, DottedDict

from .addLayer import addLayer


async def getLayer(layer: str, parameter=None, logger=False):
    # Add the Bridge
    bridge = Bridge(generateId(), logger)

    # Add the Layer
    await addLayer(bridge, layer, parameter, logger, True, True)

    # Return the Bridge
    return bridge



""" An example plugin that allows instances of `NopeRpcManager` to
    say hello. The source code can be used as a starting example.

    Please visist: nope/plugins/demo and "checkout" the examples.
"""


# @plugin("nope.dispatcher.rpcManager")
def extend(module):
    "Extends `module` - RPC-Manager"

    class NopeRpcManager(module.NopeRpcManager):
        def hello(self):
            "Ask the plugin to say hello"
            print("Hello from the Plugin")

        async def performCall(self, *args, **kwargs):
            """ Extend the original behavior """
            print("calling 'performCall' inside of our addon")
            return await module.NopeRpcManager.performCall(self, *args, **kwargs)

    def hello_func(self):
        "Ask the plugin to say hello"
        print("Hello from the Plugin")

    hello_vars = {"hello": "Hello from the Plugin"}
    return NopeRpcManager, hello_func, hello_vars
